import gym
from gym.envs.registration import register
import os

from rl_zoo3.wrappers import MaskVelocityWrapper


from RPG.envs.ant_envs import AntHEnv
from RPG.envs.modem.mj_envs.envs.hand_manipulation_suite.door_v0 import DoorEnvV0
from RPG.envs.modem.mj_envs.envs.hand_manipulation_suite.hammer_v0 import HammerEnvV0
from RPG.envs.maze import SmallMaze

from gym_donkeycar.envs.donkey_env import DonkeyEnv

try:
    import pybullet_envs  # pytype: disable=import-error
except ImportError:
    pybullet_envs = None

try:
    import highway_env  # pytype: disable=import-error
except ImportError:
    highway_env = None
else:
    # hotfix for highway_env
    import numpy as np

    np.float = np.float32  # type: ignore[attr-defined]

try:
    import neck_rl  # pytype: disable=import-error
except ImportError:
    neck_rl = None

try:
    import mocca_envs  # pytype: disable=import-error
except ImportError:
    mocca_envs = None

try:
    import custom_envs  # pytype: disable=import-error
except ImportError:
    custom_envs = None

try:
    import gym_donkeycar  # pytype: disable=import-error
except ImportError:
    gym_donkeycar = None

try:
    import panda_gym  # pytype: disable=import-error
except ImportError:
    panda_gym = None

try:
    import rocket_lander_gym  # pytype: disable=import-error
except ImportError:
    rocket_lander_gym = None

try:
    import robohive

    from robohive.envs.hands.door_v1 import DoorEnvV1
    from robohive.envs.hands.hammer_v1 import HammerEnvV1

    register(
        id="door-v1",
        entry_point="robohive.envs.hands:DoorEnvV1",
        max_episode_steps=100,
        kwargs={
            "model_path": "/home/sheelabhadra/Pi-Star/rl-baselines3-zoo/robohive/robohive/envs/hands/assets/DAPG_door.xml",
        },
    )
    register(
        id="hammer-v1",
        entry_point="robohive.envs.hands:HammerEnvV1",
        max_episode_steps=125,
        kwargs={
            "model_path": "/home/sheelabhadra/Pi-Star/rl-baselines3-zoo/robohive/robohive/envs/hands/assets/DAPG_hammer.xml",
        },
    )
except ImportError:
    robohive = None


# Register no vel envs
def create_no_vel_env(env_id: str):
    def make_env():
        env = gym.make(env_id)
        env = MaskVelocityWrapper(env)
        return env

    return make_env


for env_id in MaskVelocityWrapper.velocity_indices.keys():
    name, version = env_id.split("-v")
    register(
        id=f"{name}NoVel-v{version}",
        entry_point=create_no_vel_env(env_id),
    )

register(
    id="AntPush-v1",
    entry_point="RPG.envs.ant_envs:AntHEnv",
    max_episode_steps=400,
    kwargs={"env_name": "AntPush"},
)

register(
    id="door-v0",
    entry_point="RPG.envs.modem.mj_envs.envs.hand_manipulation_suite:DoorEnvV0",
    max_episode_steps=200,
)

register(
    id="hammer-v0",
    entry_point="RPG.envs.modem.mj_envs.envs.hand_manipulation_suite:HammerEnvV0",
    max_episode_steps=125,
)

register(id="maze2d-v0", entry_point="RPG.envs.maze:SmallMaze", max_episode_steps=20)

register(
    id="donkey-generated-roads-v0",
    entry_point="gym_donkeycar.envs.donkey_env:DonkeyEnv",
    max_episode_steps=5000,
    kwargs={"level": "generated_road"},
)
